function PPMSApplication() {
    this.hostname = window.location.hostname; // IPv4 address of the Waves System player
    this.username = 'user'; // username of the login

    this.buttonDefinition = [
        {
            id: 1,
            priority: 1,
            folders: {
                primary: 1,
                secondary: 201
            },
            feedback: true
        },
        {
            id: 2,
            priority: 2,
            folders: {
                primary: 2,
                secondary: 202
            },
            feedback: true
        },
        {
            id: 3,
            priority: 4,
            folders: {
                primary: 4
            },
            feedback: true
        },
        {
            id: 4,
            priority: 0,
            folders: {
                primary: 101
            },
            feedback: false
        }
    ];
}

PPMSApplication.prototype = {
    /**
     * Initialize the application, reading the configuration as a first step
     */
    initialize: function() {
        this.getApplicationConfiguration()
            .then(this.onGetConfigSuccess.bind(this))
            .catch(this.onGetConfigError.bind(this));
    },

    /**
     * Read the user's configuration file
     * @returns a promise resolved with the configuration file content as an object
     */
    getApplicationConfiguration: function() {
        return new Promise(function(resolve, reject) {
            var xhr = new XMLHttpRequest();
            xhr.addEventListener("readystatechange", function() {
                if(this.readyState === 4) {
                    if (xhr.status === 200) {
                        try {
                            var response = JSON.parse(this.responseText);
                            resolve(response);
                        } catch(e) {
                            reject(e);
                        }
                    } else {
                        reject("Invalid HTTP response status " + xhr.status);
                    }
                }
            });
            console.log("Getting ppms.json (you can edit this file with your settings)...");
            xhr.open("GET", "ppms.json");
            xhr.send();
        });
    },

    setBodyClass: function(clazz) {
        var elements = document.getElementsByTagName("body");
        elements[0].setAttribute("class", clazz);
    },

    onGetConfigSuccess: function(config) {
        this.hostname = config.hostname || this.hostname; // Set the user's hostname of the player
        this.username = config.username || this.username; // Set the user's username of the player
        this.initializeLogin();
    },

    onGetConfigError: function(error) {
        this.setBodyClass("init-error");

        var detail = document.getElementById("error-config-detail");
        if (detail) {
            detail.innerText = error;
        }
    },

    /**
     * Initialize the numpad and link it to the other app components
     */
     initializeLogin: function() {
        var timeout = null;
        var password = document.getElementById("password");

        this.setBodyClass("login");

        /**
         * Replace the first characters of text with bullets
         * @param {string} value is the string to process
         * @param {boolean} full is true to hide all characters, false to hide only the last one
         */
        var getBullets = function(value, full) {
            var bullet ="\u26AB";
            var bullets = "";

            if (full) {
                for (var i = 0 ; i < value.length ; i++) {
                    bullets += bullet;
                }
            } else {
                for (var i = 0 ; i < value.length  - 1 ; i++) {
                    bullets += bullet;
                }
                bullets += value.substring(value.length - 1);
            }
            return bullets;
        };

        /**
         * Set text on the password view
         * @param {string} value is the password string
         * @param {boolean} add is true when something has been added to the value
         *                      (meaning that only the last char must be displayed)
         */
        var setText = function(value, add) {
            password.textContent = getBullets(value, add);
        }

        /**
         * Callback invoked by the numpad on value changes
         * @param {string} value is the password string
         * @param {boolean} add is true when something has been added to the value
         *                      (meaning that only the last char must be displayed)
         */
        var onChange = function(value, add) {
            // Clean any pending timeout.
            // On timeout the password is totally hidden
            if (timeout != null) {
                window.clearTimeout(timeout);
                timeout = null;
            }

            if (add) {
                // Display only the last character for a short period,
                // the first characters are hidden (bullets)
                setText(value, false);
                timeout = window.setTimeout(function() {
                    setText(value, true);
                }, 300);
            } else {
                setText(value, true);
            }
        };

        this.numpad = new Numpad("numpad", {
            minlen: 4,
            maxlen: 12,
            onChange: onChange,
            onSelect: this.validatePassword.bind(this)
        });
    },

    getButtonByFolder: function(folder) {
        if (folder === undefined) {
            return null;
        }

        for (var i = 0 ; i < this.buttonDefinition.length ; i++) {
            var current = this.buttonDefinition[i];
            if (current.folders.primary === folder) {
                return current;
            }
            if (current.folders.secondary === folder) {
                return current;
            }
        }
        return null;
    },

    /**
     * Initialize the remote control UX
     **/
    initializeRemoteControl: function(ifaces) {
        var clickedButtonId;
        var Priority_Level = 7; // Priority level, from 0 (max) to 7 (min) - by default = 7 (min)

        // Show the remote control (and hide the numpad)
        this.setBodyClass("logged");

        // Get the buttons
        for (var i = 0 ; i < this.buttonDefinition.length ; i++) {
            var current = this.buttonDefinition[i];
            current.button = document.getElementById("rc-button-" + current.id);
        }

        // Given a directory, set the priority level
        var mapFolderToPriority = function(n) {
            var activeButton = this.getButtonByFolder(n);
            var isEndOfAlert = (activeButton && activeButton.folders.primary !== n) || false;

            if (isEndOfAlert) { // if current folder is "End of alert playback"
                Priority_Level = 4; // Priority_Level = 4 (min) - lower priority for PPMS
            } else if (n == 101) { // if current folder is "End of generic playback"
                Priority_Level = 7; // Priority_Level = 7 (min)
            } else if (n !== undefined) {
                Priority_Level = n; // Priority level = number of the current folder
            } else {
                Priority_Level = 7;
            }

            // Adjust the button visual
            for (var i = 0 ; i < this.buttonDefinition.length ; i++) {
                var current = this.buttonDefinition[i];
                var classList = current.button.classList;
                if (current == activeButton && activeButton.feedback) {
                    current.button.removeAttribute("disabled");
                    if (isEndOfAlert) {
                        classList.remove("rc-alert-running");
                        classList.add("rc-alert-over");
                    } else {
                        classList.add("rc-alert-running");
                        classList.remove("rc-alert-over");
                    }
                } else {
                    classList.remove("rc-alert-running");
                    classList.remove("rc-alert-over");
                    if (Priority_Level >= current.priority) {
                        current.button.removeAttribute("disabled");
                    } else {
                        current.button.setAttribute("disabled", "");
                    }
                }
            }
        }.bind(this);

        // Get the current priorty level at the open of the web page
        ifaces.player.getFolderNumber()
            .then(mapFolderToPriority);

        // Listen the player events
        ifaces.player.addListener(function (e) {
            if (e.detail.type == 'folder-enter') { // if enter folder
                mapFolderToPriority(e.detail.folder.number);
            }
        });

        var getPlaybackDelegate = function(levelMin, folder) {
            return function() {
                return ifaces.player.getFolderNumber()
                    .then(mapFolderToPriority)
                    .then(function() {
                        if (Priority_Level >= levelMin) {
                            return ifaces.player.playFolder(folder)
                                .then(function(result) {
                                    return true;
                                });
                        }
                        return false;
                    });
            };
        }

        // List of callback to be invoked on button click / confirmation
        var buttonCallbacks = [];

        for (var i = 0 ; i < this.buttonDefinition.length ; i++) {
            var current = this.buttonDefinition[i];
            var callback = getPlaybackDelegate(current.priority, current.folders.primary);
            buttonCallbacks.push(callback)
        }

        // Initialize confirmation modal
        $('#confirmDialog').on('show.bs.modal', function (event) {
            var button = $(event.relatedTarget);

            // Save the clicked button
            clickedButtonId = button.data('button');

            // Update message with the selected action
            var action = button.data('action');
            $(this).find(".modal-body span").text(action);
        });


        var thiz = this;
        $('#confirmDialog .modal-footer button').on('click', function(event) {
            var button = $(event.target);

            $(this).closest('.modal').one('hidden.bs.modal', function() {
                var id = button.data('button-id');
                if (id == 'confirmation-yes') {
                    var callback = buttonCallbacks[clickedButtonId - 1];
                    if (callback) {
                        thiz.showRemoteControlSpinnerUntil(callback())
                            .then(function(possible) {
                                if (!possible) {
                                    thiz.showErrorDialog("error-playback");
                                }
                                console.log("Play possible: " + possible);
                            })
                            .catch(function(error) {
                                thiz.showErrorDialog("error-playback");
                            });
                    }
                }
            });
        });
    },

    /**
     * Show the spinner on the remote control until a given promise
     * is resolved or rejected.
     * @param {Promise<*>} promise
     * @returns a promise
     */
    showRemoteControlSpinnerUntil: function(promise) {
        var thiz = this;
        this.setRemoteControlSpinner(true);

        return promise
            .then(function(result) {
                thiz.setRemoteControlSpinner(false);
                return result;
            })
            .catch(function(error) {
                thiz.setRemoteControlSpinner(false);
                return Promise.reject(error);
            });
    },

    /**
     * Show an error dialog.
     * @param {string} clazz Name of the class to set on the dialog body (allow displaying specific message)
     * @param {() => void} onHide callback invoked when the dialog is closed
     */
    showErrorDialog: function(clazz, onHide) {
        $("#errorDialog .modal-body").attr('class', "modal-body " + clazz);

        var dialog = $("#errorDialog");
        dialog.modal('show');

        if  (onHide) {
            var callback = function() {
                onHide();
                dialog.off('hidden.bs.modal', callback);
            }

            dialog.on('hidden.bs.modal', callback);
        }
    },

    /**
     * Validate a password, hiding the numpad and displaying the
     * remote control on success.
     * @param {string} password is the password to use
     */
    validatePassword: function(password) {
        /**
         * Callback invoked when an error dialog is closed
         */
        var onDialogClosed = function() {
            this.numpad.enableKeyboard(true);
            this.setLoginSpinner(false);
        }.bind(this);


        /*
         * Callback invoked when the login is successful
         */
        var onSuccess = function(result) {
            if (result && result.data == true) {
                this.setLoginSpinner(false);
                this.onLoginValid(password);
            } else {
                this.showErrorDialog("error-password", onDialogClosed);
            }
        }.bind(this);

        /*
         * Callback invoked on login error
         */
        var onError = function(error) {
            if (error.errors && error.errors[0]) {
                var firstError = error.errors[0];
                switch (firstError.appcode) {

                    case "NETWORK_ERROR":
                        // Cannot join the player...
                        this.showErrorDialog("error-network", onDialogClosed);
                        break;

                    case "NOT_AUTHORIZED":
                        // Invalid password
                        this.showErrorDialog("error-password", onDialogClosed);
                        break;

                    default:
                        this.showErrorDialog("error-unknown", onDialogClosed);
                        break;
                }
            } else {
                this.showErrorDialog("error-unknown", onDialogClosed);
            }
        }.bind(this);

        /**
         * Disable numpad keyboard hook and add spinner on login
         */
        this.numpad.enableKeyboard(false);
        this.setLoginSpinner(true);

        /*
         * Try authentication
         */
        var client = IDAL.getClient(this.hostname, this.username, password);
        client.post("/api/auth/login")
            .then(onSuccess)
            .catch(onError);
    },

    setLoginSpinner: function(enabled) {
        this.setSpinner("login-spinner", enabled);
    },

    setRemoteControlSpinner(enabled) {
        this.setSpinner("rc-spinner", enabled);
    },

    setSpinner: function(classname, enabled) {
        var elements = document.getElementsByClassName(classname);
        for (var i = 0 ; i < elements.length ; i++) {
            if (enabled) {
                elements[i].classList.remove("d-none");
            } else {
                elements[i].classList.add("d-none");
            }
        }
    },

    onLoginValid: function(password) {
        IDAL.getRemoteControlInterfaces(this.hostname, this.username, password)
            .then(this.initializeRemoteControl.bind(this));
    }
};



/**
 * Generic Numpad implementation.
 *
 * @param {string} id is the HTMLElement identifier of the numpad
 * @param {Object} options The options
 * @param {number} options.minlen the minimum length of the value that can be entered
 * @param {number} options.maxlen the maximum length of the value that can be entered
 * @param {(string, boolean) => void} options.onChange is the callback invoked on value change
 * @param {(string) => void} options.onSelect is the callback invoked when select is pressed
 */
function Numpad(id, options) {
    var numpad = document.getElementById(id);
    var onChange = options.onChange;
    var onSelect = options.onSelect;
    var maxlen = options.maxlen || 4;
    var minlen = options.minlen || 4;
    var value = "";

    /**
     * Add a listener on a numpad key
     * @param {string} name is the name of the key
     * @param {(Event) => void} callback is the callback to invoke when a key is clicked
     * @returns
     */
    var addClickListener = function(name, callback) {
        var key = numpad.querySelector("[name='" + name + "']");
        if (key) {
            key.addEventListener("click", callback);
        }
        return key;
    }

    /**
     * Convenience method to create a closure to add a given
     * digit to the entered value.
     *
     * @param {number} x is the digit to add to the value
     * @returns void
     */
    var getNumDelegate = function(x) {
        return function() {
            addCharToValue(x);
        }
    };

    var addCharToValue = function(x) {
        if (value.length < maxlen) {
            setValue(value + x, true);
        }
    };

    var setButtonState = function(button, enabled) {
        if (enabled) {
            button.removeAttribute("disabled");
        } else {
            button.setAttribute("disabled", "");
        }
    }

    /**
     * Set the state of the buttons depending on the input value.
     */
    var setAllButtonsState = function() {
        // Only enable OK when minlen is reached
        setButtonState(numpad_ok, value.length >= minlen);

        // Disable numpad when maxlen is reached
        for(var i = 0 ; i < numpad_numbers.length ; i++) {
            setButtonState(numpad_numbers[i], value.length < maxlen);
        }

        // Disable suppr when value is empty
        setButtonState(numpad_suppr, value.length > 0);
    };

    /**
     * Set the numpad input value
     * @param {string} v is the new value
     * @param {boolean} add is true if a digit has been added.
     */
    var setValue = function(v, add) {
        value = v;
        setAllButtonsState();
        if (onChange) {
            onChange(value, add);
        }
    };

    var mayBeSelect = function() {
        if (value.length >= minlen && onSelect) {
            onSelect(value);
        }
    };

    var backspace = function() {
        setValue(value.substring(0, value.length - 1), false);
    };

    /**
     * Add listeners on all the keys, keeping references on them to makes
     * it possible to enable or disable them.
     */
    var numpad_numbers = [];

    for (var i = 0 ; i < 10 ; i++) {
        numpad_numbers.push(addClickListener("numpad_" + i, getNumDelegate(i)));
    }

    var numpad_suppr = addClickListener("numpad_suppr", backspace);

    var numpad_ok = addClickListener("numpad_ok", function() {
        mayBeSelect();
        setValue("", false);
    });

    this._onKeyDown = function(e) {
        console.dir(e);
        switch(e.key) {
            case "0":
            case "1":
            case "2":
            case "3":
            case "4":
            case "5":
            case "6":
            case "7":
            case "8":
            case "9":
                addCharToValue(e.key);
                break;
            case "Enter":
                mayBeSelect();
                break;
            case "Backspace":
                backspace();
                break;
        }
    };
    this.enableKeyboard(true);

    setValue("", false);
}

Numpad.prototype = {
    enableKeyboard: function(enabled) {
        this.keyboard = enabled;
        if (enabled) {
            document.addEventListener("keydown", this._onKeyDown);
        } else {
            document.removeEventListener("keydown", this._onKeyDown);
        }
    }
}




// Initialize the program at the loading
var gPPMSApplication = new PPMSApplication();
window.addEventListener("load", function() {
    gPPMSApplication.initialize();
});
